# -*- Gap -*- 

LRSInfo:=NewInfoClass("LRS");
SetInfoLevel(LRSInfo,1);

LRSDefaultTimeout:=10;
LRSLongTimeout:=600;



LRS:=function()
	
    local output,LRSProcess;
    LRSProcess:= InputOutputLocalProcess(DirectoryCurrent(),
                         Filename(DirectoriesPackagePrograms("lrslib"),"lrsserv"),[]);
    if (LRSProcess=fail) then
	Info(InfoWarning,1, "Failed to start lrsserv");
        return fail;
    else	
        if (Expect(LRSProcess,"HELLO")) then
            return LRSProcess;
        else
            return fail;
        fi;	
    fi;
end;	



      
LRSLoadMatrix:=function(p,M,equalityList)
    local  rowType,i,m,n;

    m:=Size(M);
    n:=Size(M[1]);
    
    SendLine(p,m," ",n," init");
    Expect(p,"OK");
    for i in [1..m] do
        SendList(p,M[i]);
        if i in equalityList then 
            rowType:=0;
        else
            rowType:=1;
        fi;
        SendLine(p,i,rowType,"set_row");
    od;
end;

LRSSetObjective:=function(p,v,maximize)
    local dir,j;
    
    if maximize then 
        dir:=1;
    else
        dir:=0;
    fi;
    SendList(p,v);
    SendLine(p,dir,"set_obj");
end;

    
LRSSolveLP:=function(p)
    local output;
    SendLine(p,"solve_lp");
    return Expect(p,"OK");
end;


LRSGetInt:=function(p)
    local line,vec;
    line:=ReadLineWithTimeout(p,LRSLongTimeout);
    if (line=fail) then 
        return fail;
    fi;
    Chomp(line);
    NormalizeWhitespace(line);
    return Int(line);
end;

LRSGetVec:=function(p)
    local l;
    l:=ReadLineWithTimeout(p);
    if (l=fail) then
        return fail;
    fi;
    return StringToRatVec(l);
end;

       
# note, at the moment this does not shift the rhs back to column zero.
# that is, it is a direct analogue of lrs_getsolution.  Caveat usator    
LRSGetSolution:=function(p,col)
    SendLine(p,col,"getsolution");
    if (Expect(p,"OK")) then
        return LRSGetVec(p);
    else
        return fail;
    fi;
end;

LRSGetFirstBasis:=function(p)
    local dim;
    SendLine(p,"getfirstbasis");
    dim:=LRSGetInt(p);
    if (dim <> fail) then
        return rec(dim:=dim);
    else
        return fail;
    fi;
end;

LRSGetNextBasis:=function(p)
    SendLine(p,"getnextbasis");
    return Expect(p,"OK",LRSLongTimeout);
end;

LRSSetVRepresentation:=function(p)
    SendLine(p,"v_rep");
end;

LRSSetDebug:=function(p)
    SendLine(p,"debug");
end;

LRSSetGeometric:=function(p)
    SendLine(p,"geometric");
end;

LRSGetLinearities:=function(p)
    local M,rows,j;
    M:=[];
    SendLine(p,"getlinearities");
    if (Expect(p,"OK")) then
        rows:=LRSGetInt(p);
        for j in [1..rows] do
            Add(M,LRSGetVec(p));
        od;
    fi;
    return M;
end;

LRSGetDict:=function(p)
    local Det,B,Row,C,Col,A,line,list;
    
    A:=[];
    
    SendLine(p,"get_dict");
    Det:=LRSGetInt(p);
    B:=LRSGetVec(p);
    Row:=LRSGetVec(p);
    C:=LRSGetVec(p);    
    Col:=LRSGetVec(p);
    
    repeat
        line:=ReadLineWithTimeout(p,LRSLongTimeout);
        Info(LRSInfo,3,"lrslib returned ",line);
        if line=fail then 
            return fail;
        fi;
        line:=Chomp(line);
        if (line <> "DONE") then
            list:=StringToRatVec(line);
            Add(A,list);
        fi;
    until line="DONE";
    
    return rec(Det:=Det, B:=B, Row:=Row, C:=C, Col:=Col, A:=A);
end;
    
LRSGetCobasis:=function(p,col)
    local Det,Cobasis,Extraincidence,Ray,Totalincidence;
    
    SendLine(p,col,"getcobasis");
    Det:=LRSGetInt(p);
    Ray:=LRSGetInt(p);
    Cobasis:=LRSGetVec(p);
    Totalincidence:=LRSGetInt(p);
    Extraincidence:=LRSGetVec(p);
    return rec(det:=Det,
               ray:=Ray,
               orderedcobasis:=Cobasis,
               cobasis:=AsSet(Cobasis),
               totalincidence:=Totalincidence,
               extraincidence:=Extraincidence);
end;

LRSSetCobasis:=function(p,cob)
    SendList(p,cob);
    SendLine(p,"setcobasis");
    return(Expect(p,"OK",LRSLongTimeout));
end;
LRSGetVertex:=function(p)
    local coords;
    
    SendLine(p,"getvertex");
    coords:=LRSGetVec(p);
    return coords;
end;


LRSAllRatio:=function(p,facet)
    local Pivots;
    SendLine(p,facet,"allratio");
    Pivots:=LRSGetVec(p);
    return Pivots;
end;


LRSOrbitRatio:=function(p,facet)
    local Pivots;
    SendLine(p,facet,"orbitratio");
    Pivots:=LRSGetVec(p);
    return Pivots;
end;

LRSLexSigns:=function(p,list)
    SendItems(p,"@");
    SendList(p,list);
    SendLine(p,"set_lex_signs");
end;

LRSLexRatio:=function(p,facet)
    local enter;
    SendLine(p,facet,"lexratio");
    if (not Expect(p,"OK")) then
        return fail;
    else
        enter:=LRSGetInt(p);
        return enter;
    fi;
end;

LRSPivot:=function(p,leave,enter)
    SendLine(p,leave,enter,"pivot");
    return(Expect(p,"OK"));
end;

LRSDumpDict:=function(p)
    SendLine(p,"dump_dict");
end;

LRSDumpOrbits:=function(p)
    SendLine(p,"dump_orbits");
end;

LRSSetOptions:=function(l,options)
    local names;
    names:=RecNames(options);
    
    if "VRepresentation" in names and options.VRepresentation then
        LRSSetVRepresentation(l);
    fi;
    if "debug" in names and options.debug then
        LRSSetDebug(l);
    fi;
end;

LRSOrbits:=function(l,orbits)
    local orb,oo;
    SendItems(l,"@");
    
    oo:=ShallowCopy(orbits);
    Sort(oo);
    for orb in Reversed(oo) do
        SendItems(l,"@");
        SendList(l,orb);
    od;
    SendLine(l,"set_orbits");

end;
               
      
          