
RequirePackage("Nauty");

InfoGram:=NewInfoClass("Gram");
SetInfoLevel(InfoGram,1);

InnerProdMat:=function(cob,M,InnerProduct)
    return List(M{cob},u->List(M{cob},v->InnerProduct(u,v)));
end;



#  loops are omitted on the premise that all vectors will have same norm
gramEdges:=function(cob,M)
    local i,j,adj,row;
    
    adj:=[];
    
    
    for i in [1..Size(cob)] do
        for j in [i+1..Size(cob)] do
            if  M[cob[i]][cob[j]]<>0 then
                AddSet(adj,[cob[i],cob[j]]);
            fi;
        od;
    od;
    return adj;
end;


#  loops are omitted on the premise that all vectors will have same norm
colouredGramEdges:=function(i,M)
    local j,colour,adj,lists,lastCol,index,edge;
    
    adj:=[];
    
    Info(InfoGram,2,"Scanning for non-zero inner products");
    for j in [1..Size(M)] do
        if j <> i then
            colour:= M[i][j];

            if  colour <> 0 then
                Add(adj,[colour,[i,j]]);
            fi;
        fi;
    od;

    Info(InfoGram,2,"Sorting labelled edges");
    Sort(adj);
    lists:=[[]];
    
    Info(InfoGram,2,"Building Lists");
    lastCol:=adj[1][1];
    index:=1;
    for edge in adj do
        if edge[1] <> lastCol then
            index:=index+1;
            lists[index]:=[];
            lastCol:=edge[1];
        fi;
        Add(lists[index],edge[2]);
    od;
    return lists;
end;

colouredGramEdges2:=function(M)
    local i,j,colour,adj,lists,lastCol,index,edge;
    
    adj:=[];
    
    Info(InfoGram,2,"Scanning for non-zero inner products");
    for i in [1..Size(M)] do
        for j in [i+1..Size(M)] do
            colour:= M[i][j];
            if  colour <> 0 then
                Add(adj,[colour,[i,j]]);
            fi;
        od;
    od;
    
    Info(InfoGram,3,"Raw adjacencies=",adj);
    Info(InfoGram,2,"Sorting labelled edges");
    Sort(adj);
    lists:=[[]];
    
    Info(InfoGram,2,"Building Lists");
    lastCol:=adj[1][1];
    index:=1;
    for edge in adj do
        if edge[1] <> lastCol then
            index:=index+1;
            lists[index]:=[];
            lastCol:=edge[1];
        fi;
        Add(lists[index],edge[2]);
    od;
    Info(InfoGram,3,"Edge classes=",lists);
    return lists;
end;

basicGramGraph:=function(cobasis,M)
    
    local vertices,edges,m,adj,i,d,e,colours,cob,vertexColours,edgeColours,
          uniqueEdgeColours,offset,indices,pi,nu;
    
    
    d:=Size(cobasis);
    vertices:=[1..d];
    
    edges:=gramEdges(vertices,M);
    
    
    m:=Size(vertices)+Size(edges);
    
    
#    if  IsList(stabilizerOrbits) then
#        vertexColours:= List(cobasis,
#                             j->First([1..Size(stabilizerOrbits)],
#                                     i-> j in stabilizerOrbits[i]));
#        vertexColours:= List(vertexColours,j->Size(stabilizerOrbits[j]));
#    else
    vertexColours:= List(cobasis, j->0);    
#    fi;

    
    pi:=Sortex(vertexColours);
    nu:=pi^(-1);
    
    offset:=Maximum(vertexColours);
    
    Info(InfoGram,4,"edges = ",edges);
    edges:=List(edges,e->OnSets(e,pi));
    Info(InfoGram,4,"new edges = ",edges);    
    Info(InfoGram,4,"pi = ",pi);    
    edgeColours:=List([1..Size(edges)], 
                      j->M[OnPoints(edges[j][1],nu)][OnPoints(edges[j][2],nu)]);

    
    uniqueEdgeColours:=AsSet(edgeColours);
    Info(InfoGram,4,"edgeColours = ",edgeColours);
    edgeColours:=List(edgeColours,c->Position(uniqueEdgeColours,c)+offset);
    Info(InfoGram,4,"new edgeColours = ",edgeColours);
    
    adj:=List([1..m],p->[]);
    
    for i in [1..Size(edges)] do
        e:=edges[i];
        AddSet(adj[e[1]],i+d);
        AddSet(adj[i+d],e[1]);        
        AddSet(adj[e[2]],i+d);
        AddSet(adj[i+d],e[2]);        
    od;
    

    

    colours:=Concatenation(vertexColours,edgeColours);
    Info(InfoGram,4,"colours = ",colours);
    return rec(numVerts:=d+Size(edges), adjList:=adj, colours:=colours,
               labels:=Permuted(cobasis,pi));
end;

makeAdjList:=function(m,edges)
    local adj,e;
    
    adj:=List([1..m],p->[]);
    
    for e in edges do
        AddSet(adj[e[1]],e[2]);        
        AddSet(adj[e[2]],e[1]);        
    od;
    return adj;
end;

buildGramCopyGraph:=function(M)
    local E,i,j,offset,pair,L,d;
    
    E:=colouredGramEdges2(M);
    L:=[];
    d:=Size(M);
    
    for i in [1..Size(E)] do
        offset:=(i-1)*d;
        if (i>1) then
            # is is sufficient just to connect to the previous copy
            # or do we need a clique?
            for j in [1..d] do
                Add(L,[j+(i-2)*d,j+offset]);
            od;
        fi;

        for pair in E[i] do 
            Add(L,[pair[1]+offset,pair[2]+offset]);
        od;
    od;
    Info(InfoGram,3,"copied graph=",L);
    return rec(numVerts:=d*Size(E), 
               adjList:=makeAdjList(d*Size(E),L), 
               colours:=List([1..d*Size(E)],p->1),
               labels:=List([1..d*Size(E)]));
end;



buildGramGraph:=function(cobasis,M)
    
    local vertices,edges,m,adj,i,d,e,colours,cob,vertexColours,edgeColours,
          uniqueEdgeColours,offset,indices,pi,nu;
    
    
    d:=Size(cobasis);
    vertices:=[1..d];
    
    
    edges:=gramEdges(vertices,M);
    
    
    m:=Size(vertices)+Size(edges);
    
    
#    if  IsList(stabilizerOrbits) then
#        vertexColours:= List(cobasis,
#                             j->First([1..Size(stabilizerOrbits)],
#                                     i-> j in stabilizerOrbits[i]));
#        vertexColours:= List(vertexColours,j->Size(stabilizerOrbits[j]));
#    else
    vertexColours:= List(cobasis, j->0);    
#    fi;

    
    pi:=Sortex(vertexColours);
    nu:=pi^(-1);
    
    offset:=Maximum(vertexColours);
    
    Info(InfoGram,4,"edges = ",edges);
    edges:=List(edges,e->OnSets(e,pi));
    Info(InfoGram,4,"new edges = ",edges);    
    Info(InfoGram,4,"pi = ",pi);    
    edgeColours:=List([1..Size(edges)], 
                      j->M[OnPoints(edges[j][1],nu)][OnPoints(edges[j][2],nu)]);

    
    uniqueEdgeColours:=AsSet(edgeColours);
    Info(InfoGram,4,"edgeColours = ",edgeColours);
    edgeColours:=List(edgeColours,c->Position(uniqueEdgeColours,c)+offset);
    Info(InfoGram,4,"new edgeColours = ",edgeColours);
    
    
    adj:=List([1..m],p->[]);
    
    for i in [1..Size(edges)] do
        e:=edges[i];
        AddSet(adj[e[1]],i+d);
        AddSet(adj[i+d],e[1]);        
        AddSet(adj[e[2]],i+d);
        AddSet(adj[i+d],e[2]);        
    od;
    

    colours:=Concatenation(vertexColours,edgeColours);
    Info(InfoGram,4,"colours = ",colours);
    return rec(numVerts:=d+Size(edges), adjList:=adj, colours:=colours,
               labels:=Permuted(cobasis,pi));
end;

gramMat:=function(cob,M,InnerProduct)
    local G,B, pi,L,L2;
    
    G:=InnerProdMat(cob,M,InnerProduct);
#    G:=B*TransposedMat(B);
    
#    Info(InfoGram,2,"raw= ",G,"\n");
#    L:=[1..Size(G)];
#    Sort(L,function(i,j) return SortedList(G[i]) < SortedList(G[j]); end);
#    pi:=PermList(L);
 #   Info(InfoGram,2,"perm = ",pi,"\n");
    #    return OnGraphMat(G,pi);
    return G;
end;

DefaultInnerProduct:=function(u,v)
    return u*v;
end;

gramVec:=function(coblist,M,InnerProduct)
    local G,B, L;
    G:=InnerProdMat(coblist,M,InnerProduct);
    L:=(AsSortedList(List(G,AsSortedList))); 
    return L;
end;    


fastGramMat:=function(coblist,Mat)
    return List(coblist,i->List(coblist,j->Mat[i][j]));
end;

fastGramVec:=function(coblist,Mat)
    local G,B, L;
    G:=fastGramMat(coblist,Mat);
    return(AsSortedList(List(G,AsSortedList))); 
end;    

canonGramViaNauty:=function(G,cob,ident,nauty)
    local gg,p,automorphs,rgens,labelPerm, tmpLabels,j;
    
    gg:=NautyGraph(G.numVerts,G.adjList,G.colours,ident);
    
    NautyLoadGraph(nauty,gg);
    automorphs:=NautyFindAutomorphisms(nauty);
    p:=NautyGetCanon(nauty);
        # we actually want the perm to the canonical labelling.
    p:=p^(-1);

    gg.colours:=Permuted(gg.colours,p);
        # note that the permutation will not necessarily make sense 
        # restricted to vertices
    gg.perm:=p;
    
    gg.labels:=G.labels{Filtered(Permuted([1..gg.numVertices],p), j->j<=Size(cob))};
    
    # We are computing something like a "local stabilizer"; it must 
    # the stabilizer actions, but it may be more symmetric at a single basis
    # than the polyhedron as a whole.
    
    Info(InfoGram,4,"Nauty perm  ",gg.ident,"=",p);
    Info(InfoGram,4,"raw automorphs ",automorphs);
    if (Size(automorphs)>0) then
        rgens:=List(automorphs,
                    pi->
                    MappingPermListList(G.labels,
                            Permuted(G.labels,
                                    RestrictedPerm(pi,[1..Size(cob)]))));
    
        gg.automorphismGroup:=Group(rgens);
        Info(InfoGram,4,"gram group ", gg.automorphismGroup);
    else
        gg.automorphismGroup:=Group(());
    fi;
    
    
    gg.adjList:=List(Permuted(gg.adjList,p),L->OnSets(L,p));
    return gg;
end;


groupViaGram:=function(Mat,InnerProduct,nauty)
    local gM,gg;
    
    gM:=InnerProdMat([1..Size(Mat)],Mat,InnerProduct);
    gg:=buildGramGraph([1..Size(Mat)],gM);
    
    gg:=canonGramViaNauty(gg,[1..Size(Mat)],0, nauty);
    return gg.automorphismGroup;
end;

groupViaGram2:=function(Mat,InnerProduct)
    local gM,ge,S,L,i;
    
    Info(InfoGram,1,"Computing Angles");        
    gM:=InnerProdMat([1..Size(Mat)],Mat,InnerProduct);
    
    S:=SymmetricGroup(Size(Mat));
    
    for i in [1..Size(Mat)] do
                
        Info(InfoGram,1,"Angles with ", i);    
        ge:=colouredGramEdges(i,gM);
   
        for L in ge do
            Info(InfoGram,1,"Stabilizer order=", Order(S));
            S:=Stabilizer(S,L,OnSetsSets);
        od;
    od;
    return S;
         
end;
    
    
groupViaGram3:=function(Mat,InnerProduct,nauty)
    local GM,graph,gg;
    
    GM:=gramMat([1..Size(Mat)],Mat,InnerProduct);
    graph:=buildGramCopyGraph(GM);
    
    gg:=canonGramViaNauty(graph,[1..Size(Mat)],0, nauty);
    return Group(List(GeneratorsOfGroup(gg.automorphismGroup),
                   pi->RestrictedPerm(pi,[1..Size(Mat)])));
end;
    
